from collections import defaultdict
import os
import threading
import time as ttime
import warnings

import numpy as np
from bec_lib import MessageEndpoints, bec_logger, messages
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus, Kind
from ophyd import DynamicDeviceComponent as Dcpt
from ophyd import OphydObject, PositionerBase, Signal
from ophyd.sim import EnumSignal, SynSignal
from ophyd.utils import LimitError, ReadOnlyError
from ophyd_devices.sim.sim_data import SimulatedDataBase, SimulatedDataCamera, SimulatedDataMonitor

from ophyd_devices.sim.sim_signals import SetableSignal, ReadOnlySignal, ComputedReadOnlySignal

logger = bec_logger.logger


class DeviceStop(Exception):
    pass


class SimMonitor(Device):
    """
    A simulated device mimic any 1D Axis (position, temperature, beam).

    Readback functionality can be configured

    Parameters
    ----------
    name : string, keyword only
    value : object, optional
        The initial value. Default is 0.
    delay : number, optional
        Simulates how long it takes the device to "move". Default is 0 seconds.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    USER_ACCESS = ["sim"]

    sim_cls = SimulatedDataMonitor

    readback = Cpt(ComputedReadOnlySignal, value=0, kind=Kind.hinted)

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        *,
        name,
        value=0,
        delay=0,
        precision=3,
        tolerance: float = 0.5,
        sim_init: dict = None,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        self.delay = delay
        self.precision = precision
        self.tolerance = tolerance
        self.init_sim_params = sim_init
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name


class SynGaussBEC(Device):
    """
    Evaluate a point on a Gaussian based on the value of a motor.

    Parameters
    ----------
    name : string
    motor : Device
    motor_field : string
    center : number
        center of peak
    Imax : number
        max intensity of peak
    sigma : number, optional
        Default is 1.
    noise : {'poisson', 'uniform', None}, optional
        Add noise to the gaussian peak.
    noise_multiplier : float, optional
        Only relevant for 'uniform' noise. Multiply the random amount of
        noise by 'noise_multiplier'
    random_state : numpy random state object, optional
        np.random.RandomState(0), to generate random number with given seed

    Example
    -------
    motor = SynAxis(name='motor')
    det = SynGauss('det', motor, 'motor', center=0, Imax=1, sigma=1)
    """

    val = Cpt(ComputedReadOnlySignal, value=0, kind=Kind.hinted)
    Imax = Cpt(Signal, value=10, kind=Kind.config)
    center = Cpt(Signal, value=0, kind=Kind.config)
    sigma = Cpt(Signal, value=1, kind=Kind.config)
    motor = Cpt(Signal, value="samx", kind=Kind.config)
    noise = Cpt(
        EnumSignal,
        value="none",
        kind=Kind.config,
        enum_strings=("none", "poisson", "uniform"),
    )
    noise_multiplier = Cpt(Signal, value=1, kind=Kind.config)

    def __init__(self, name, *, device_manager=None, random_state=None, **kwargs):
        self.device_manager = device_manager
        set_later = {}
        for k in ("sigma", "noise", "noise_multiplier"):
            v = kwargs.pop(k, None)
            if v is not None:
                set_later[k] = v
        self.sim_state = defaultdict(lambda: {})
        super().__init__(name=name, **kwargs)
        self.sim_state[self.name] = self.sim_state.pop(self.val.name, None)
        self.val.name = self.name

        self.random_state = random_state or np.random
        self.precision = 3

        for k, v in set_later.items():
            getattr(self, k).put(v)

    def _compute_sim_state(self, signal_name: str) -> None:
        try:
            m = self.device_manager.devices[self.motor.get()].obj.read()[self.motor.get()]["value"]
            # we need to do this one at a time because
            #   - self.read() may be screwed with by the user
            #   - self.get() would cause infinite recursion
            Imax = self.Imax.get()
            center = self.center.get()
            sigma = self.sigma.get()
            noise = self.noise.get()
            noise_multiplier = self.noise_multiplier.get()
            v = Imax * np.exp(-((m - center) ** 2) / (2 * sigma**2))
            if noise == "poisson":
                v = int(self.random_state.poisson(np.round(v), 1))
            elif noise == "uniform":
                v += self.random_state.uniform(-1, 1) * noise_multiplier
            self.sim_state[signal_name]["value"] = v
            self.sim_state[signal_name]["timestamp"] = ttime.time()
        except Exception as exc:
            logger.warning(f"Failed to compute sim state with exception {exc}")
            self.sim_state[signal_name]["value"] = 0
            self.sim_state[signal_name]["timestamp"] = ttime.time()

    def get(self, *args, **kwargs):
        self.sim_state["readback"] = self._compute()
        self.sim_state["readback_ts"] = ttime.time()
        return self.val.get()


class _SLSDetectorConfigSignal(Signal):
    def put(self, value, *, timestamp=None, force=False):
        self._readback = value
        self.parent.sim_state[self.name] = value

    def get(self):
        self._readback = self.parent.sim_state[self.name]
        return self.parent.sim_state[self.name]


class SimCamera(Device):
    USER_ACCESS = ["sim"]

    sim_cls = SimulatedDataCamera
    SHAPE = (100, 100)

    SUB_MONITOR = "monitor"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_path = Cpt(SetableSignal, name="file_path", value="", kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)
    save_file = Cpt(SetableSignal, name="save_file", value=False, kind=Kind.config)

    # image shape, only adjustable via config.
    image_shape = Cpt(SetableSignal, name="image_shape", value=SHAPE, kind=Kind.config)
    image = Cpt(
        ComputedReadOnlySignal,
        name="image",
        value=np.empty(SHAPE, dtype=np.uint16),
        kind=Kind.omitted,
    )

    def __init__(
        self,
        *,
        name,
        kind=None,
        parent=None,
        sim_init: dict = None,
        device_manager=None,
        **kwargs,
    ):
        self.device_manager = device_manager
        self.init_sim_params = sim_init
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self._stopped = False
        self.file_name = ""
        self.metadata = {}

    def trigger(self):
        status = DeviceStatus(self)

        self.subscribe(status._finished, event_type=self.SUB_ACQ_DONE, run=False)

        def acquire():
            try:
                for _ in range(self.burst.get()):
                    # Send data for each trigger
                    self._run_subs(sub_type=self.SUB_MONITOR, value=self.image.get())
                    if self._stopped:
                        raise DeviceStop
            except DeviceStop:
                pass
            finally:
                self._stopped = False
                self._done_acquiring()

        threading.Thread(target=acquire, daemon=True).start()
        return status

    def stage(self) -> list[object]:
        """Stage the camera

        Receive scan message from REDIS first, extract relevant scan data,
        and set all signals for the scan, e.g. scan_number, file_name, frames, etc.
        """
        msg = self.device_manager.producer.get(MessageEndpoints.scan_status())
        scan_msg = messages.ScanStatusMessage.loads(msg)
        self.metadata = {
            "scanID": scan_msg.content["scanID"],
            "RID": scan_msg.content["info"]["RID"],
            "queueID": scan_msg.content["info"]["queueID"],
        }
        scan_number = scan_msg.content["info"]["scan_number"]
        self.frames.set(scan_msg.content["info"]["num_points"])
        self.file_name = os.path.join(
            self.file_path.get(), self.file_pattern.get().format(scan_number)
        )
        self._stopped = False
        return super().stage()

    def _send_data_to_bec(self) -> None:
        config_readout = {
            signal.item.name: signal.item.get()
            for signal in self.walk_signals()
            if signal.item._kind == Kind.config
        }

        signals = {"config": config_readout, "data": self.file_name}
        msg = messages.DeviceMessage(signals=signals, metadata=self.metadata)
        self.device_manager.producer.set_and_publish(
            MessageEndpoints.device_read(self.name), msg.dumps()
        )

    def unstage(self) -> list[object]:
        """Unstage the device

        Send reads from all config signals to redis
        """
        if self._stopped is True or not self._staged:
            return super().unstage()
        self._send_data_to_bec()

        return super().unstage()

    def stop(self, *, success=False):
        self._stopped = True
        super().stop(success=success)


class DummyController:
    USER_ACCESS = [
        "some_var",
        "controller_show_all",
        "_func_with_args",
        "_func_with_args_and_kwargs",
        "_func_with_kwargs",
        "_func_without_args_kwargs",
    ]
    some_var = 10
    another_var = 20

    def on(self):
        self._connected = True

    def off(self):
        self._connected = False

    def _func_with_args(self, *args):
        return args

    def _func_with_args_and_kwargs(self, *args, **kwargs):
        return args, kwargs

    def _func_with_kwargs(self, **kwargs):
        return kwargs

    def _func_without_args_kwargs(self):
        return None

    def controller_show_all(self):
        """dummy controller show all

        Raises:
            in: _description_
            LimitError: _description_

        Returns:
            _type_: _description_
        """
        print(self.some_var)


class DummyControllerDevice(Device):
    USER_ACCESS = ["controller"]


class SynFlyer(Device, PositionerBase):
    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_manager = device_manager

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)

    @property
    def hints(self):
        return {"fields": ["flyer_samx", "flyer_samy"]}

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            self.name: {
                                "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                                "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                            }
                        },
                        metadata={"pointID": ii, **metadata},
                    ).dumps()
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    device.device_manager.producer.send(
                        MessageEndpoints.device_read(device.name), bundle.dumps()
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.producer.set_and_publish(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name,
                            status=1,
                            metadata={"pointID": ii, **metadata},
                        ).dumps(),
                    )
            device.device_manager.producer.send(
                MessageEndpoints.device_read(device.name), bundle.dumps()
            )
            device.device_manager.producer.set_and_publish(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name,
                    status=0,
                    metadata={"pointID": num_pos, **metadata},
                ).dumps(),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()


class SynController(OphydObject):
    def on(self):
        pass

    def off(self):
        pass


class SynFlyerLamNI(Device, PositionerBase):
    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_manager = device_manager

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.controller = SynController(name="SynController")

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            "syn_flyer_lamni": {
                                "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                                "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                            }
                        },
                        metadata={"pointID": ii, **metadata},
                    ).dumps()
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    device.device_manager.producer.send(
                        MessageEndpoints.device_read(device.name), bundle.dumps()
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.producer.set_and_publish(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name,
                            status=1,
                            metadata={"pointID": ii, **metadata},
                        ).dumps(),
                    )
            device.device_manager.producer.send(
                MessageEndpoints.device_read(device.name), bundle.dumps()
            )
            device.device_manager.producer.set_and_publish(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name,
                    status=0,
                    metadata={"pointID": num_pos, **metadata},
                ).dumps(),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()


class SimPositioner(Device, PositionerBase):
    """
    A simulated device mimicing any 1D Axis device (position, temperature, rotation).

    Parameters
    ----------
    name : string, keyword only
    readback_func : callable, optional
        When the Device is set to ``x``, its readback will be updated to
        ``f(x)``. This can be used to introduce random noise or a systematic
        offset.
        Expected signature: ``f(x) -> value``.
    value : object, optional
        The initial value. Default is 0.
    delay : number, optional
        Simulates how long it takes the device to "move". Default is 0 seconds.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    # Specify which attributes are accessible via BEC client
    USER_ACCESS = ["sim", "readback", "speed", "dummy_controller"]

    sim_cls = SimulatedDataBase

    # Define the signals as class attributes
    readback = Cpt(ReadOnlySignal, name="readback", value=0, kind=Kind.hinted)
    setpoint = Cpt(SetableSignal, value=0, kind=Kind.normal)
    motor_is_moving = Cpt(SetableSignal, value=0, kind=Kind.normal)

    # Config signals
    velocity = Cpt(SetableSignal, value=1, kind=Kind.config)
    acceleration = Cpt(SetableSignal, value=1, kind=Kind.config)

    # Ommitted signals
    high_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    low_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    unused = Cpt(Signal, value=1, kind=Kind.omitted)

    # TODO add short description to these two lines and explain what this does
    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=1,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        limits=None,
        tolerance: float = 0.5,
        sim: dict = None,
        **kwargs,
    ):
        # Whether motions should be instantaneous or depend on motor velocity
        self.delay = delay
        self.precision = precision
        self.tolerance = tolerance
        self.init_sim_params = sim

        self.speed = speed
        self.update_frequency = update_frequency
        self._stopped = False
        self.dummy_controller = DummyController()

        # initialize inner dictionary with simulated state
        self.sim = self.sim_cls(parent=self, **kwargs)

        super().__init__(name=name, labels=labels, kind=kind, **kwargs)
        # Rename self.readback.name to self.name, also in self.sim_state
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name
        # Init limits from deviceConfig
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        """Return the limits of the simulated device."""
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        """Return the low limit of the simulated device."""
        return self.limits[0]

    @property
    def high_limit(self):
        """Return the high limit of the simulated device."""
        return self.limits[1]

    def check_value(self, value: any):
        """
        Check that requested position is within existing limits.

        This function has to be implemented on the top level of the positioner.
        """
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not low_limit <= value <= high_limit:
            raise LimitError(f"position={value} not within limits {self.limits}")

    def _set_sim_state(self, signal_name: str, value: any) -> None:
        """Update the simulated state of the device."""
        self.sim.sim_state[signal_name]["value"] = value
        self.sim.sim_state[signal_name]["timestamp"] = ttime.time()

    def _get_sim_state(self, signal_name: str) -> any:
        """Return the simulated state of the device."""
        return self.sim.sim_state[signal_name]["value"]

    def move(self, value: float, **kwargs) -> DeviceStatus:
        """Change the setpoint of the simulated device, and simultaneously initiated a motion."""
        self._stopped = False
        self.check_value(value)
        old_setpoint = self._get_sim_state(self.setpoint.name)
        self._set_sim_state(self.motor_is_moving.name, 1)
        self._set_sim_state(self.setpoint.name, value)

        def update_state(val):
            """Update the state of the simulated device."""
            if self._stopped:
                raise DeviceStop
            old_readback = self._get_sim_state(self.readback.name)
            self._set_sim_state(self.readback.name, val)

            # Run subscription on "readback"
            self._run_subs(
                sub_type=self.SUB_READBACK,
                old_value=old_readback,
                value=self.sim.sim_state[self.readback.name]["value"],
                timestamp=self.sim.sim_state[self.readback.name]["timestamp"],
            )

        st = DeviceStatus(device=self)
        if self.delay:
            # If self.delay is not 0, we use the speed and updated frequency of the device to compute the motion
            def move_and_finish():
                """Move the simulated device and finish the motion."""
                success = True
                try:
                    # Compute final position with some jitter
                    move_val = self._get_sim_state(
                        self.setpoint.name
                    ) + self.tolerance * np.random.uniform(-1, 1)
                    # Compute the number of updates needed to reach the final position with the given speed
                    updates = np.ceil(
                        np.abs(old_setpoint - move_val) / self.speed * self.update_frequency
                    )
                    # Loop over the updates and update the state of the simulated device
                    for ii in np.linspace(old_setpoint, move_val, int(updates)):
                        ttime.sleep(1 / self.update_frequency)
                        update_state(ii)
                    # Update the state of the simulated device to the final position
                    update_state(move_val)
                    self._set_sim_state(self.motor_is_moving, 0)
                except DeviceStop:
                    success = False
                finally:
                    self._stopped = False
                # Call function from positioner base to indicate that motion finished with success
                self._done_moving(success=success)
                # Set status to finished
                st.set_finished()

            # Start motion in Thread
            threading.Thread(target=move_and_finish, daemon=True).start()

        else:
            # If self.delay is 0, we move the simulated device instantaneously
            update_state(value)
            self._done_moving()
            st.set_finished()
        return st

    def stop(self, *, success=False):
        """Stop the motion of the simulated device."""
        super().stop(success=success)
        self._stopped = True

    @property
    def position(self):
        """Return the current position of the simulated device."""
        return self.readback.get()

    @property
    def egu(self):
        """Return the engineering units of the simulated device."""
        return "mm"


class SynDynamicComponents(Device):
    messages = Dcpt({f"message{i}": (SynSignal, None, {"name": f"msg{i}"}) for i in range(1, 6)})


class SynDeviceSubOPAAS(Device):
    zsub = Cpt(SimPositioner, name="zsub")


class SynDeviceOPAAS(Device):
    x = Cpt(SimPositioner, name="x")
    y = Cpt(SimPositioner, name="y")
    z = Cpt(SynDeviceSubOPAAS, name="z")


if __name__ == "__main__":
    gauss = SynGaussBEC(name="gauss")



from abc import ABC, abstractmethod
from collections import defaultdict
import enum
import time as ttime
import numpy as np

from bec_lib import bec_logger

logger = bec_logger.logger


class SimulatedDataException(Exception):
    """Exception raised when there is an issue with the simulated data."""


class SimulationType(str, enum.Enum):
    """Type of simulation to steer simulated data."""

    CONSTANT = "constant"
    GAUSSIAN = "gauss"


class NoiseType(str, enum.Enum):
    """Type of noise to add to simulated data."""

    NONE = "none"
    UNIFORM = "uniform"
    POISSON = "poisson"


class SimulatedDataBase:
    USER_ACCESS = [
        "get_sim_params",
        "set_sim_params",
        "get_sim_type",
        "set_sim_type",
    ]

    def __init__(self, *args, parent=None, device_manager=None, **kwargs) -> None:
        self.parent = parent
        self.sim_state = defaultdict(lambda: {})
        self._all_params = defaultdict(lambda: {})
        self.device_manager = device_manager
        self._simulation_type = None
        self.init_paramaters(**kwargs)
        self._active_params = self._all_params.get(self._simulation_type, None)

    def init_paramaters(self, **kwargs):
        """Initialize the parameters for the Simulated Data

        This methods should be implemented by the subclass.

        It sets the default parameters for the simulated data in
        self._params and calls self._update_init_params()
        """

    def get_sim_params(self) -> dict:
        """Return the currently parameters for the active simulation type in sim_type.

        These parameters can be changed with set_sim_params.

        Returns:
            dict: Parameters of the currently active simulation in sim_type.
        """
        return self._active_params

    def set_sim_params(self, params: dict) -> None:
        """Change the current set of parameters for the active simulation type.

        Args:
            params (dict): New parameters for the active simulation type.

        Raises:
            SimulatedDataException: If the new parameters can not be set or is not part of the parameters initiated.
        """
        for k, v in params.items():
            try:
                if k == "noise":
                    self._active_params[k] = NoiseType(v)
                else:
                    self._active_params[k] = v
            except Exception as exc:
                raise SimulatedDataException(
                    f"Could not set {k} to {v} in {self._active_params} with exception {exc}"
                ) from exc

    def get_sim_type(self) -> SimulationType:
        """Return the simulation type of the simulation.

        Returns:
            SimulationType: Type of simulation (e.g. "constant" or "gauss).
        """
        return self._simulation_type

    def set_sim_type(self, simulation_type: SimulationType) -> None:
        """Set the simulation type of the simulation."""
        try:
            self._simulation_type = SimulationType(simulation_type)
        except ValueError as exc:
            raise SimulatedDataException(
                f"Could not set simulation type to {simulation_type}. Valid options are 'constant'"
                " and 'gauss'"
            ) from exc
        self._active_params = self._all_params.get(self._simulation_type, None)

    def _compute_sim_state(self, signal_name: str) -> None:
        """Update the simulated state of the device.

        If no computation is relevant, ignore this method.
        Otherwise implement it in the subclass.
        """

    def update_sim_state(self, signal_name: str, value: any) -> None:
        """Update the simulated state of the device.

        Args:
            signal_name (str): Name of the signal to update.
        """
        self.sim_state[signal_name]["value"] = value
        self.sim_state[signal_name]["timestamp"] = ttime.time()

    def _update_init_params(self, sim_type_default: SimulationType) -> None:
        """Update the initial parameters of the simulated data with input from deviceConfig.

        Args:
            sim_type_default (SimulationType): Default simulation type to use if not specified in deviceConfig.
        """
        init_params = self.parent.init_sim_params
        for sim_type in self._all_params.values():
            for sim_type_config_element in sim_type:
                if init_params:
                    if sim_type_config_element in init_params:
                        sim_type[sim_type_config_element] = init_params[sim_type_config_element]
        # Set simulation type to default if not specified in deviceConfig
        sim_type_select = (
            init_params.get("sim_type", sim_type_default) if init_params else sim_type_default
        )
        self.set_sim_type(sim_type_select)


class SimulatedDataMonitor(SimulatedDataBase):
    """Simulated data for a monitor."""

    def init_paramaters(self, **kwargs):
        """Initialize the parameters for the simulated data

        This method will fill self._all_params with the default parameters for
        SimulationType.CONSTANT and SimulationType.GAUSSIAN.
        New simulation types can be added by adding a new key to self._all_params,
        together with the required parameters for that simulation type. Please
        also complement the docstring of this method with the new simulation type.

        For SimulationType.CONSTANT:
            Amp is the amplitude of the constant value.
            Noise is the type of noise to add to the signal. Available options are 'poisson', 'uniform' or 'none'.
            Noise multiplier is the multiplier of the noise, only relevant for uniform noise.

        For SimulationType.GAUSSIAN:
            ref_motor is the motor that is used as reference to compute the gaussian.
            amp is the amplitude of the gaussian.
            cen is the center of the gaussian.
            sig is the sigma of the gaussian.
            noise is the type of noise to add to the signal. Available options are 'poisson', 'uniform' or 'none'.
            noise multiplier is the multiplier of the noise, only relevant for uniform noise.
        """
        self._all_params = {
            SimulationType.CONSTANT: {
                "amp": 100,
                "noise": NoiseType.POISSON,
                "noise_multiplier": 0.1,
            },
            SimulationType.GAUSSIAN: {
                "ref_motor": "samx",
                "amp": 100,
                "cen": 0,
                "sig": 1,
                "noise": NoiseType.NONE,
                "noise_multiplier": 0.1,
            },
        }
        # Update init parameters and set simulation type to Constant if not specified otherwise in init_sim_params
        self._update_init_params(sim_type_default=SimulationType.CONSTANT)

    def _compute_sim_state(self, signal_name: str) -> None:
        """Update the simulated state of the device.

        It will update the value in self.sim_state with the value computed by
        the chosen simulation type.

        Args:
            signal_name (str): Name of the signal to update.
        """
        if self.get_sim_type() == SimulationType.CONSTANT:
            value = self._compute_constant()
        elif self.get_sim_type() == SimulationType.GAUSSIAN:
            value = self._compute_gaussian()

        self.update_sim_state(signal_name, value)

    def _compute_constant(self) -> float:
        """Computes constant value and adds noise if activated."""
        v = self._active_params["amp"]
        if self._active_params["noise"] == NoiseType.POISSON:
            v = np.random.poisson(np.round(v), 1)[0]
            return v
        elif self._active_params["noise"] == NoiseType.UNIFORM:
            v += np.random.uniform(-1, 1) * self._active_params["noise_multiplier"]
            return v
        elif self._active_params["noise"] == NoiseType.NONE:
            v = self._active_params["amp"]
            return v
        else:
            # TODO Propagate msg to client!
            logger.warning(
                f"Unknown noise type {self._active_params['noise']}. Please choose from 'poisson',"
                " 'uniform' or 'none'. Returning 0."
            )
            return 0

    def _compute_gaussian(self) -> float:
        """Computes return value for sim_type = "gauss".

        The value is based on the parameters for the gaussian in
        self._active_params and the position of the ref_motor
        and adds noise based on the noise type.

        If computation fails, it returns 0.

        Returns: float
        """

        params = self._active_params
        try:
            motor_pos = self.device_manager.devices[params["ref_motor"]].obj.read()[
                params["ref_motor"]
            ]["value"]
            v = params["amp"] * np.exp(
                -((motor_pos - params["cen"]) ** 2) / (2 * params["sig"] ** 2)
            )
            if params["noise"] == NoiseType.POISSON:
                v = np.random.poisson(np.round(v), 1)[0]
            elif params["noise"] == NoiseType.UNIFORM:
                v += np.random.uniform(-1, 1) * params["noise_multiplier"]
            return v
        except SimulatedDataException as exc:
            # TODO Propagate msg to client!
            logger.warning(
                f"Could not compute gaussian for {params['ref_motor']} with {exc} raised."
                "Returning 0 instead."
            )
            return 0


class SimulatedDataCamera(SimulatedDataBase):
    """Simulated class to compute data for a 2D camera."""

    def init_paramaters(self, **kwargs):
        """Initialize the parameters for the simulated data

        This method will fill self._all_params with the default parameters for
        SimulationType.CONSTANT and SimulationType.GAUSSIAN.
        New simulation types can be added by adding a new key to self._all_params,
        together with the required parameters for that simulation type. Please
        also complement the docstring of this method with the new simulation type.

        For SimulationType.CONSTANT:
            Amp is the amplitude of the constant value.
            Noise is the type of noise to add to the signal. Available options are 'poisson', 'uniform' or 'none'.
            Noise multiplier is the multiplier of the noise, only relevant for uniform noise.

        For SimulationType.GAUSSIAN:
            amp is the amplitude of the gaussian.
            cen_off is the pixel offset from the center of the gaussian from the center of the image.
                It is passed as a numpy array.
            cov is the 2D covariance matrix used to specify the shape of the gaussian.
                It is a 2x2 matrix and will be passed as a numpy array.
            noise is the type of noise to add to the signal. Available options are 'poisson', 'uniform' or 'none'.
            noise multiplier is the multiplier of the noise, only relevant for uniform noise.
        """
        self._all_params = {
            SimulationType.CONSTANT: {
                "amp": 100,
                "noise": NoiseType.POISSON,
                "noise_multiplier": 0.1,
            },
            SimulationType.GAUSSIAN: {
                "amp": 100,
                "cen_off": np.array([0, 0]),
                "cov": np.array([[10, 5], [5, 10]]),
                "noise": NoiseType.NONE,
                "noise_multiplier": 0.1,
            },
        }
        # Update init parameters and set simulation type to Gaussian if not specified otherwise in init_sim_params
        self._update_init_params(sim_type_default=SimulationType.GAUSSIAN)

    def _compute_sim_state(self, signal_name: str) -> None:
        """Update the simulated state of the device.

        It will update the value in self.sim_state with the value computed by
        the chosen simulation type.

        Args:
            signal_name (str): Name of the signal to update.
        """
        if self.get_sim_type() == SimulationType.CONSTANT:
            value = self._compute_constant()
        elif self.get_sim_type() == SimulationType.GAUSSIAN:
            value = self._compute_gaussian()

        self.update_sim_state(signal_name, value)

    def _compute_constant(self) -> float:
        """Compute a return value for sim_type = Constant."""

        # tuple with shape
        shape = self.sim_state[self.parent.image_shape.name]["value"]
        v = self._active_params["amp"] * np.ones(shape, dtype=np.uint16)
        if self._active_params["noise"] == NoiseType.POISSON:
            v = np.random.poisson(np.round(v), v.shape)
            return v
        if self._active_params["noise"] == NoiseType.UNIFORM:
            multiplier = self._active_params["noise_multiplier"]
            v += np.random.randint(-multiplier, multiplier, v.shape)
            return v
        if self._active_params["noise"] == NoiseType.NONE:
            return v
        # TODO Propagate msg to client!
        logger.warning(
            f"Unknown noise type {self._active_params['noise']}. Please choose from 'poisson',"
            " 'uniform' or 'none'. Returning 0."
        )
        return 0

    def _compute_multivariate_gaussian(
        self, pos: np.ndarray, cen_off: np.ndarray, cov: np.ndarray
    ) -> np.ndarray:
        """Computes and returns the multivariate Gaussian distribution.

        Args:
            pos (np.ndarray): Position of the gaussian.
            cen_off (np.ndarray): Offset from cener of image for the gaussian.
            cov (np.ndarray): Covariance matrix of the gaussian.

        Returns:
            np.ndarray: Multivariate Gaussian distribution.
        """

        dim = cen_off.shape[0]
        cov_det = np.linalg.det(cov)
        cov_inv = np.linalg.inv(cov)
        N = np.sqrt((2 * np.pi) ** dim * cov_det)
        # This einsum call calculates (x-mu)T.Sigma-1.(x-mu) in a vectorized
        # way across all the input variables.
        fac = np.einsum("...k,kl,...l->...", pos - cen_off, cov_inv, pos - cen_off)

        return np.exp(-fac / 2) / N

    def _compute_gaussian(self) -> float:
        """Computes return value for sim_type = "gauss".

        The value is based on the parameters for the gaussian in
        self._active_params and adds noise based on the noise type.

        If computation fails, it returns 0.

        Returns: float
        """

        params = self._active_params
        shape = self.sim_state[self.parent.image_shape.name]["value"]
        try:
            X, Y = np.meshgrid(
                np.linspace(-shape[0] / 2, shape[0] / 2, shape[0]),
                np.linspace(-shape[1] / 2, shape[1] / 2, shape[1]),
            )
            pos = np.empty((*X.shape, 2))
            pos[:, :, 0] = X
            pos[:, :, 1] = Y

            v = self._compute_multivariate_gaussian(
                pos=pos, cen_off=params["cen_off"], cov=params["cov"]
            )
            # divide by max(v) to ensure that maximum is params["amp"]
            v *= params["amp"] / np.max(v)

            # TODO add dependency from motor position -> #transmission factor, sigmoidal form from 0 to 1 as a function of motor pos
            # motor_pos = self.device_manager.devices[params["ref_motor"]].obj.read()[
            #     params["ref_motor"]
            # ]["value"]

            if params["noise"] == NoiseType.POISSON:
                v = np.random.poisson(np.round(v), v.shape)
                return v
            if params["noise"] == NoiseType.UNIFORM:
                multiplier = params["noise_multiplier"]
                v += np.random.uniform(-multiplier, multiplier, v.shape)
                return v
            if self._active_params["noise"] == NoiseType.NONE:
                return v
        except SimulatedDataException as exc:
            # TODO Propagate msg to client!
            logger.warning(
                f"Could not compute gaussian for {params['ref_motor']} with {exc} raised."
                "Returning 0 instead."
            )
            return 0



from ophyd import Device, Component, EpicsMotor, PseudoPositioner, PseudoSingle
from ophyd.pseudopos import pseudo_position_argument, real_position_argument


class SlitH(PseudoPositioner):
    """Python wrapper for virtual slits

    These devices should be implemented as an EPICS SoftMotor IOC,
    but thats not the case for all slits. So here is a pure ophyd
    implementation. Uses standard naming convention!

    NOTE: The real and virtual axes are wrapped together.
    """

    # Motor interface
    x1 = Component(EpicsMotor, "TRX1")
    x2 = Component(EpicsMotor, "TRX2")

    cenx = Component(PseudoSingle)
    gapx = Component(PseudoSingle)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        return self.RealPosition(
            x1=pseudo_pos.cenx - pseudo_pos.gapx / 2, x2=pseudo_pos.cenx + pseudo_pos.gapx / 2
        )

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(
            cenx=(real_pos.x1 + real_pos.x2) / 2, gapx=real_pos.x2 - real_pos.x1
        )


class SlitV(PseudoPositioner):
    """Python wrapper for virtual slits

    These devices should be implemented as an EPICS SoftMotor IOC,
    but thats not the case for all slits. So here is a pure ophyd
    implementation. Uses standard naming convention!

    NOTE: The real and virtual axes are wrapped together.
    """

    # Motor interface
    y1 = Component(EpicsMotor, "TRY1")
    y2 = Component(EpicsMotor, "TRY2")

    ceny = Component(PseudoSingle)
    gapy = Component(PseudoSingle)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        return self.RealPosition(
            y1=pseudo_pos.ceny - pseudo_pos.gapy / 2, y2=pseudo_pos.ceny + pseudo_pos.gapy / 2
        )

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(
            ceny=(real_pos.y1 + real_pos.y2) / 2, gapy=real_pos.y2 - real_pos.y1
        )



SMARACT_ERRORS = {
    1: "Syntax Error: The command could not be processed due to a syntactical error.",
    2: "Invalid Command Error: The command given is not known to the system.",
    3: "Overflow Error: This error occurs if a parameter given is too large and therefore cannot be processed.",
    4: "Parse Error: The command could not be processed due to a parse error.",
    5: "Too Few Parameters Error: The specified command requires more parameters in order to be executed.",
    6: "Too Many Parameters Error: There were too many parameters given for the specified command.",
    7: "Invalid Parameter Error: A parameter given exceeds the valid range. Please see the command description for valid ranges of the parameters.",
    8: "Wrong Mode Error: This error is generated if the specified command is not available in the current communication mode. For example, the SRC command is not executable in synchronous mode.",
    129: "No Sensor Present Error: This error occurs if a command was given that requires sensor feedback, but the addressed positioner has none attached.",
    140: "Sensor Disabled Error: This error occurs if a command was given that requires sensor feedback, but the sensor of the addressed positioner is disabled (see SSE command).",
    141: "Command Overridden Error: This error is only generated in the asynchronous communication mode. When the software commands a movement which is then interrupted by the Hand Control Module, an error of this type is generated.",
    142: "End Stop Reached Error: This error is generated in asynchronous mode if the target position of a closed-loop command could not be reached, because a mechanical end stop was detected. After this error the positioner will have a movement status code of 0 (stopped).",
    143: "Wrong Sensor Type Error:  This error occurs if a closed-loop command does not match the sensor type that is currently configured for the addressed channel. For example, issuing a GP command while the targeted channel is configured as rotary will lead to this error.",
    144: "Could Not Find Reference Mark Error: This error is generated in asynchronous mode (see SCM) if the search for a reference mark was aborted.",
    145: "Wrong End Effector Type Error: This error occurs if a command does not match the end effector type that is currently configured for the addressed channel. For example, sending GF while the targeted channel is configured for a gripper will lead to this error.",
    146: "Movement Locked Error: This error occurs if a movement command is issued while the system is in the locked state. ",
    147: "Range Limit Reached Error: If a range limit is defined (SPL or SAL) and the positioner is about to move beyond this limit, then the positioner will stop and report this error (only in asynchronous mode, see SCM). After this error the positioner will have status code of 0 (stopped).",
    148: "Physical Position Unknown Error: A range limit is only allowed to be defined if the positioner “knows” its physical position. If this is not the case, the commands SPL and SAL will return this error code.",
    150: "Command Not Processable Error: This error is generated if a command is sent to a channel when it is in a state where the command cannot be processed. For example, to change the sensor type of a channel the addressed channel must be completely stopped. In this case send a stop command before changing the type.",
    151: "Waiting For Trigger Error: If there is at least one command queued in the command queue then you may only append more commands (if the queue is not full), but you may not issue movement commands for immediate execution. Doing so will generate this error. See section 2.4.5 “Command Queues“.",
    152: "Command Not Triggerable Error: After sending a ATC command you are required to issue a movement command that is to be triggered by the given event source. Commands that cannot be triggered will generate this error.",
    153: "Command Queue Full Error: This error is generated if you attempt to append more commands to the command queue, but the queue cannot hold anymore commands. The queue capacity may be read out with a get channel property command (GCP on p.30).",
    154: "Invalid Component Error: Indicates that a component (e.g. SCP) was selected that does not exist.",
    155: "Invalid Sub Component Error: Indicates that a sub component (e.g. SCP) was selected that does not exist.",
    156: "Invalid Property Error: Indicates that a property (e.g. SCP) was selected that does not exist.",
    157: "Permission Denied Error: This error is generated when you call a functionality which is not unlocked for the system (e.g. Low Vibration Mode).",
}


class SmaractError(Exception):
    pass


class SmaractCommunicationError(SmaractError):
    pass


class SmaractErrorCode(SmaractError):
    def __init__(self, error_code: int, message=""):
        self.error_code = error_code
        self.error_code_message = SMARACT_ERRORS.get(error_code, "UNKNOWN ERROR")
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.error_code} / {self.error_code_message}. {self.message}"



from .eiger1p5m_csaxs.eiger1p5m import Eiger1p5MDetector
from .epics import *
from .galil.galil_ophyd import GalilMotor
from .galil.sgalil_ophyd import SGalilMotor
from .npoint.npoint import NPointAxis
from .rt_lamni import RtLamniMotor
from .sim.sim import SimPositioner, SimMonitor
from .sim.sim import SimPositioner as SynAxisOPAAS
from .sim.sim import SimMonitor as SynAxisMonitor
from .sim.sim import SimMonitor as SynGaussBEC
from .sim.sim import SimCamera as SynSLSDetector
from .sim.sim_signals import ReadOnlySignal as SynSignalRO
from .sim.sim import SynDeviceOPAAS, SynFlyer
from .sls_devices.sls_devices import SLSInfo, SLSOperatorMessages
from .smaract.smaract_ophyd import SmaractMotor
from .utils.static_device_test import launch



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 17:06:51 2021

@author: mohacsi_i
"""

import os
import yaml
from ophyd.ophydobj import OphydObject

from devices import *

# ####################################################
# Test connection to beamline devices
# ####################################################
# Current file path
path = os.path.dirname(os.path.abspath(__file__))

# Load simulated device database
fp = open(f"{path}/db/test_database.yml", "r")
lut_db = yaml.load(fp, Loader=yaml.Loader)

# Load beamline specific database
bl = os.getenv("BEAMLINE_XNAME", "TESTENV")
if bl != "TESTENV":
    fp = open(f"{path}/db/{bl.lower()}_database.yml", "r")
    lut_db.update(yaml.load(fp, Loader=yaml.Loader))


def createProxy(name: str, connect=True) -> OphydObject:
    """Create an ophyd device from its alias

    Factory routine that uses the beamline database to instantiate
    ophyd devices from their alias and pre-defined configuration.
    Does nothing if the device is already an OphydObject!
    """
    if issubclass(type(name), OphydObject):
        return name

    entry = lut_db[name]
    # Yeah, using global namespace
    cls_candidate = globals()[entry["type"]]
    # print(f"Class candidate: {cls_candidate}")

    if issubclass(cls_candidate, OphydObject):
        ret = cls_candidate(**entry["config"])
        if connect:
            ret.wait_for_connection(timeout=5)
        return ret
    else:
        raise RuntimeError(f"Unsupported return class: {entry['type']}")


if __name__ == "__main__":
    num_errors = 0
    for key in lut_db:
        try:
            dut = createProxy(str(key))
            print(f"{key}\t: {type(dut)}\t{dut.read()}")
            # print(f"{key}\t: {type(dut)}")
        except Exception as ex:
            num_errors += 1
            print(key)
            print(ex)
    print(f"\nTotal number of errors: {num_errors}")



import logging
import threading
import time

import numpy as np
from bec_lib import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, PositionerBase, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError, ReadOnlyError

from ophyd_devices.smaract.smaract_controller import SmaractController
from ophyd_devices.smaract.smaract_errors import SmaractCommunicationError, SmaractError
from ophyd_devices.utils.controller import threadlocked
from ophyd_devices.utils.socket import SocketIO, SocketSignal, raise_if_disconnected

logger = bec_logger.logger


class SmaractSignalBase(SocketSignal):
    def __init__(self, signal_name, **kwargs):
        self.signal_name = signal_name
        super().__init__(**kwargs)
        self.controller = self.parent.controller
        self.sock = self.parent.controller.sock


class SmaractSignalRO(SmaractSignalBase):
    def __init__(self, signal_name, **kwargs):
        super().__init__(signal_name, **kwargs)
        self._metadata["write_access"] = False

    @threadlocked
    def _socket_set(self, val):
        raise ReadOnlyError("Read-only signals cannot be set")


class SmaractReadbackSignal(SmaractSignalRO):
    @threadlocked
    def _socket_get(self):
        return self.controller.get_position(self.parent.axis_Id_numeric) * self.parent.sign


class SmaractSetpointSignal(SmaractSignalBase):
    setpoint = 0

    @threadlocked
    def _socket_get(self):
        return self.setpoint

    @threadlocked
    def _socket_set(self, val):
        target_val = val * self.parent.sign
        self.setpoint = target_val

        if self.controller.axis_is_referenced(self.parent.axis_Id_numeric):
            self.controller.move_axis_to_absolute_position(self.parent.axis_Id_numeric, target_val)
            # parameters are axis_no,pos_mm*1e6, hold_time_sec*1e3
        else:
            raise SmaractError(f"Axis {self.parent.axis_Id_numeric} is not referenced.")


class SmaractMotorIsMoving(SmaractSignalRO):
    @threadlocked
    def _socket_get(self):
        return self.controller.is_axis_moving(self.parent.axis_Id_numeric)


class SmaractAxisReferenced(SmaractSignalRO):
    @threadlocked
    def _socket_get(self):
        return self.parent.controller.axis_is_referenced(self.parent.axis_Id_numeric)


class SmaractAxisLimits(SmaractSignalBase):
    @threadlocked
    def _socket_get(self):
        limits_msg = self.controller.socket_put_and_receive(f"GPL{self.parent.axis_Id_numeric}")
        if limits_msg.startswith(":PL"):
            limits = [
                float(limit)
                for limit in limits_msg.strip(f":PL{self.parent.axis_Id_numeric},").split(",")
            ]
        else:
            raise SmaractCommunicationError("Expected to receive message starting with :PL.")
        return limits

    # def _socket_set(self, val):


class SmaractMotor(Device, PositionerBase):
    USER_ACCESS = ["controller"]
    readback = Cpt(
        SmaractReadbackSignal,
        signal_name="readback",
        kind="hinted",
    )
    user_setpoint = Cpt(SmaractSetpointSignal, signal_name="setpoint")

    # motor_resolution = Cpt(
    #    SmaractMotorResolution, signal_name="resolution", kind="config"
    # )

    motor_is_moving = Cpt(SmaractMotorIsMoving, signal_name="motor_is_moving", kind="normal")
    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")
    # all_axes_referenced = Cpt(
    #    SmaractAxesReferenced, signal_name="all_axes_referenced", kind="config"
    # )

    SUB_READBACK = "readback"
    SUB_CONNECTION_CHANGE = "connection_change"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        axis_Id,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        host="mpc2680.psi.ch",
        port=8085,
        limits=None,
        sign=1,
        socket_cls=SocketIO,
        **kwargs,
    ):
        self.controller = SmaractController(
            socket_cls=socket_cls, socket_host=host, socket_port=port
        )
        self.axis_Id = axis_Id
        self.sign = sign
        self.controller.set_axis(axis=self, axis_nr=self.axis_Id_numeric)
        self.tolerance = kwargs.pop("tolerance", 0.5)

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name
        self.controller.subscribe(
            self._update_connection_state, event_type=self.SUB_CONNECTION_CHANGE
        )
        self._update_connection_state()
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def _update_connection_state(self, **kwargs):
        for walk in self.walk_signals():
            walk.item._metadata["connected"] = self.controller.connected

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False
        timeout = kwargs.pop("timeout", 4)
        status = super().move(position, timeout=timeout, **kwargs)
        self.user_setpoint.put(position, wait=False)

        def move_and_finish():
            while self.motor_is_moving.get():
                val = self.readback.read()
                self._run_subs(
                    sub_type=self.SUB_READBACK,
                    value=val,
                    timestamp=time.time(),
                )
                time.sleep(0.1)
            val = self.readback.read()
            success = np.isclose(
                val[self.name]["value"],
                position,
                atol=self.tolerance,
            )
            self._done_moving(success=success)

        threading.Thread(target=move_and_finish, daemon=True).start()
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    def axis_Id(self):
        return self._axis_Id_alpha

    @axis_Id.setter
    def axis_Id(self, val):
        if isinstance(val, str):
            if len(val) != 1:
                raise ValueError(f"Only single-character axis_Ids are supported.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = ord(val.lower()) - 97
        else:
            raise TypeError(f"Expected value of type str but received {type(val)}")

    @property
    def axis_Id_numeric(self):
        return self._axis_Id_numeric

    @axis_Id_numeric.setter
    def axis_Id_numeric(self, val):
        if isinstance(val, int):
            if val > 26:
                raise ValueError(f"Numeric value exceeds supported range.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = (chr(val + 97)).capitalize()
        else:
            raise TypeError(f"Expected value of type int but received {type(val)}")

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return "mm"

    def stage(self) -> list[object]:
        return super().stage()

    def unstage(self) -> list[object]:
        return super().unstage()

    def stop(self, *, success=False):
        self.controller.stop_all_axes()
        return super().stop(success=success)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    mock = False
    if not mock:
        lsmarA = SmaractMotor("A", name="lsmarA", host="mpc2680.psi.ch", port=8085, sign=1)
        lsmarB = SmaractMotor("B", name="lsmarB", host="mpc2680.psi.ch", port=8085, sign=1)

        lsmarA.stage()
        lsmarB.stage()
        # status = leyey.move(2, wait=True)
        # status = leyey.move(2, wait=True)
        lsmarA.read()
        lsmarB.read()

        lsmarA.get()
        lsmarB.get()
        lsmarA.describe()

        lsmarA.unstage()
        lsmarA.controller.off()
        # status = leyey.move(10, wait=False)
        # print(lSmaract_controller)
    else:
        from ophyd_devices.utils.socket import SocketMock

        lsmarA = SmaractMotor(
            "A",
            name="lsmarA",
            host="mpc2680.psi.ch",
            port=8085,
            sign=1,
            socket_cls=SocketMock,
        )
        lsmarB = SmaractMotor(
            "B",
            name="lsmarB",
            host="mpc2680.psi.ch",
            port=8085,
            sign=1,
            socket_cls=SocketMock,
        )
        lsmarA.stage()
        lsmarB.stage()

        lsmarA.read()
        lsmarB.read()

        lsmarA.get()
        lsmarB.get()
        lsmarA.describe()

        lsmarA.unstage()
        lsmarA.controller.off()
        # status = leyey.move(10, wait=False)
        # print(lSmaract_controller)
